#ifndef SENSOR_IMU_H
#define SENSOR_IMU_H

//wolf includes
#include "core/sensor/sensor_base.h"
#include "core/utils/params_server.hpp"
#include <iostream>

namespace wolf {

WOLF_STRUCT_PTR_TYPEDEFS(ParamsSensorImu);


struct ParamsSensorImu : public ParamsSensorBase
{
    //noise std dev
    double w_noise = 0.001; //standard deviation of Gyroscope noise (same for all the axis) in rad/sec/ sqrt(s)
    double a_noise = 0.04; //standard deviation of Acceleration noise (same for all the axis) in m/s2/sqrt(s)

    //Initial biases std dev
    double ab_initial_stdev = 0.01; //accelerometer micro_g/sec
    double wb_initial_stdev = 0.01; //gyroscope rad/sec

    // bias rate of change std dev
    double ab_rate_stdev = 0.00001;
    double wb_rate_stdev = 0.00001;

    virtual ~ParamsSensorImu() = default;
    ParamsSensorImu()
    {
        //DEFINED FOR COMPATIBILITY PURPOSES. TO BE REMOVED IN THE FUTURE.
    }
    ParamsSensorImu(std::string _unique_name, const ParamsServer& _server):
        ParamsSensorBase(_unique_name, _server)
    {
        w_noise             = _server.getParam<double>(prefix + _unique_name + "/w_noise");
        a_noise             = _server.getParam<double>(prefix + _unique_name + "/a_noise");
        ab_initial_stdev    = _server.getParam<double>(prefix + _unique_name + "/ab_initial_stdev");
        wb_initial_stdev    = _server.getParam<double>(prefix + _unique_name + "/wb_initial_stdev");
        ab_rate_stdev       = _server.getParam<double>(prefix + _unique_name + "/ab_rate_stdev");
        wb_rate_stdev       = _server.getParam<double>(prefix + _unique_name + "/wb_rate_stdev");
    }
    std::string print() const
    {
        return "\n" + ParamsSensorBase::print()                           + "\n"
            + "w_noise: "           + std::to_string(w_noise)           + "\n"
            + "a_noise: "           + std::to_string(a_noise)           + "\n"
            + "ab_initial_stdev: "  + std::to_string(ab_initial_stdev)  + "\n"
            + "wb_initial_stdev: "  + std::to_string(wb_initial_stdev)  + "\n"
            + "ab_rate_stdev: "     + std::to_string(ab_rate_stdev)     + "\n"
            + "wb_rate_stdev: "     + std::to_string(wb_rate_stdev)     + "\n";
    }
};

WOLF_PTR_TYPEDEFS(SensorImu);

class SensorImu : public SensorBase
{

    protected:
        double a_noise; //Power Spectral Density (same for all the axis) in micro_g/ sqrt(Hz)
        double w_noise; //Rate Noise Spectral Density (same for all the axis) in deg/sec/ sqrt(Hz)

        //This is a trial to factor how much can the bias change in 1 sec at most
        double ab_initial_stdev; //accelerometer m/sec^s
        double wb_initial_stdev; //gyroscope rad/sec
        double ab_rate_stdev;    //accelerometer m/sec^2 / sqrt(sec)
        double wb_rate_stdev;    //gyroscope rad/sec / sqrt(sec)

    public:

        SensorImu(const Eigen::VectorXd& _extrinsics, const ParamsSensorImu& _params);
        SensorImu(const Eigen::VectorXd& _extrinsics, ParamsSensorImuPtr _params);
        WOLF_SENSOR_CREATE(SensorImu, ParamsSensorImu, 7);

        double getGyroNoise() const;
        double getAccelNoise() const;
        double getWbInitialStdev() const;
        double getAbInitialStdev() const;
        double getWbRateStdev() const;
        double getAbRateStdev() const;

        virtual ~SensorImu();

};

inline double SensorImu::getGyroNoise() const
{
    return w_noise;
}

inline double SensorImu::getAccelNoise() const
{
    return a_noise;
}

inline double SensorImu::getWbInitialStdev() const
{
    return wb_initial_stdev;
}

inline double SensorImu::getAbInitialStdev() const
{
    return ab_initial_stdev;
}

inline double SensorImu::getWbRateStdev() const
{
    return wb_rate_stdev;
}

inline double SensorImu::getAbRateStdev() const
{
    return ab_rate_stdev;
}

} // namespace wolf

#endif // SENSOR_Imu_H
