//Wolf
#include "imu/capture/capture_imu.h"
#include "imu/processor/processor_imu.h"
#include "imu/sensor/sensor_imu.h"
#include "core/common/wolf.h"
#include "core/problem/problem.h"
#include "core/state_block/state_block.h"
#include "core/state_block/state_quaternion.h"
#include <core/utils/utils_gtest.h>
#include "core/utils/logging.h"

class FeatureImu_test : public testing::Test
{

    public: //These can be accessed in fixtures
        wolf::ProblemPtr problem;
        wolf::TimeStamp ts;
        wolf::CaptureImuPtr imu_ptr;
        Eigen::VectorXd state_vec;
        Eigen::VectorXd delta_preint;
        Eigen::Matrix<double,9,9> delta_preint_cov;
        std::shared_ptr<wolf::FeatureImu> feat_imu;
        wolf::FrameBasePtr last_frame;
        wolf::FrameBasePtr origin_frame;
        Eigen::MatrixXd dD_db_jacobians;
        wolf::ProcessorBasePtr processor_ptr_;
        wolf::ProcessorMotionPtr processor_motion_ptr_;

    //a new of this will be created for each new test
    virtual void SetUp()
    {
        using namespace wolf;
        using std::shared_ptr;
        using std::make_shared;
        using std::static_pointer_cast;

        // Wolf problem
        problem = Problem::create("POV", 3);
        Eigen::VectorXd Imu_extrinsics(7);
        Imu_extrinsics << 0,0,0, 0,0,0,1; // Imu pose in the robot
        ParamsSensorImuPtr sen_imu_params = std::make_shared<ParamsSensorImu>();
        SensorBasePtr sensor_ptr = problem->installSensor("SensorImu", "Main Imu", Imu_extrinsics, sen_imu_params);
        ProcessorParamsImuPtr prc_imu_params = std::make_shared<ProcessorParamsImu>();
        prc_imu_params->max_time_span   = 0.5;
        prc_imu_params->max_buff_length = 10;
        prc_imu_params->dist_traveled   = 5;
        prc_imu_params->angle_turned    = 0.5;
        processor_ptr_ = problem->installProcessor("ProcessorImu", "Imu pre-integrator", sensor_ptr, prc_imu_params);
        processor_motion_ptr_ = std::static_pointer_cast<ProcessorMotion>(processor_ptr_);

    // Time and data variables
        TimeStamp t;
        Eigen::Vector6d data_;

        t.set(0);

    // Set the origin
        Eigen::VectorXd x0(10);
        x0 << 0,0,0,  0,0,0,1,  0,0,0;
        MatrixXd P0; P0.setIdentity(9,9);
        origin_frame = problem->setPrior(x0, P0, 0.0, 0.01);

    // Emplace one capture to store the Imu data arriving from (sensor / callback / file / etc.)
    // give the capture a big covariance, otherwise it will be so small that it won't pass following assertions
        imu_ptr = std::static_pointer_cast<CaptureImu>(
                CaptureBase::emplace<CaptureImu>(origin_frame,
                                                 t,
                                                 sensor_ptr,
                                                 data_,
                                                 Eigen::Matrix6d::Identity(),
                                                 Eigen::Vector6d::Zero()) );

    //process data
        data_ << 2, 0, 9.8, 0, 0, 0;
        t.set(0.1);
        // Expected state after one integration
        //x << 0.01,0,0, 0,0,0,1, 0.2,0,0; // advanced at a=2m/s2 during 0.1s ==> dx = 0.5*2*0.1^2 = 0.01; dvx = 2*0.1 = 0.2
    // assign data to capture
        imu_ptr->setData(data_);
        imu_ptr->setTimeStamp(t);
    // process data in capture
        sensor_ptr->process(imu_ptr);

    //emplace Frame
        problem->getCurrentStateAndStamp(state_vec, ts);
        last_frame = problem->emplaceFrame(KEY, state_vec, ts);

    //emplace a feature
        delta_preint            = processor_motion_ptr_->getMotion().delta_integr_;
        delta_preint_cov        = processor_motion_ptr_->getMotion().delta_integr_cov_ + MatrixXd::Identity(9,9)*1e-08;
        VectorXd calib_preint   = processor_motion_ptr_->getLast()->getCalibrationPreint();
        dD_db_jacobians         = processor_motion_ptr_->getMotion().jacobian_calib_;
        feat_imu                = std::static_pointer_cast<FeatureImu>(
                FeatureBase::emplace<FeatureImu>(imu_ptr,
                                                 delta_preint,
                                                 delta_preint_cov,
                                                 calib_preint,
                                                 dD_db_jacobians,
                                                 imu_ptr) );
    }

    virtual void TearDown()
    {
        // code here will be called just after the test completes
        // ok to through exceptions from here if need be
        /*
            You can do deallocation of resources in TearDown or the destructor routine.
                However, if you want exception handling you must do it only in the TearDown code because throwing an exception
                from the destructor results in undefined behavior.
            The Google assertion macros may throw exceptions in platforms where they are enabled in future releases.
                Therefore, it's a good idea to use assertion macros in the TearDown code for better maintenance.
        */
    }
};

TEST_F(FeatureImu_test, check_frame)
{
    // set variables
    using namespace wolf;

    FrameBasePtr left_frame = feat_imu->getFrame();
    wolf::TimeStamp t;
    left_frame->getTimeStamp(t);
    origin_frame->getTimeStamp(ts);

    ASSERT_NEAR(t.get(), ts.get(), wolf::Constants::EPS_SMALL) << "t != ts \t t=" << t << "\t ts=" << ts << std::endl;
    ASSERT_TRUE(origin_frame->isKey());
    ASSERT_TRUE(last_frame->isKey());
    ASSERT_TRUE(left_frame->isKey());

    wolf::StateBlockPtr origin_pptr, origin_optr, origin_vptr, left_pptr, left_optr, left_vptr;
    origin_pptr = origin_frame->getP();
    origin_optr = origin_frame->getO();
    origin_vptr = origin_frame->getV();
    left_pptr = left_frame->getP();
    left_optr = left_frame->getO();
    left_vptr = left_frame->getV();

    ASSERT_MATRIX_APPROX(origin_pptr->getState(), left_pptr->getState(), wolf::Constants::EPS_SMALL);
    Eigen::Map<const Eigen::Quaterniond> origin_Quat(origin_optr->getState().data()), left_Quat(left_optr->getState().data());
    ASSERT_QUATERNION_APPROX(origin_Quat, left_Quat, wolf::Constants::EPS_SMALL);
    ASSERT_MATRIX_APPROX(origin_vptr->getState(), left_vptr->getState(), wolf::Constants::EPS_SMALL);

    ASSERT_EQ(origin_frame->id(), left_frame->id());
}

TEST_F(FeatureImu_test, access_members)
{
    using namespace wolf;

    Eigen::VectorXd delta(10);
    //dx = 0.5*2*0.1^2 = 0.01; dvx = 2*0.1 = 0.2; dz = 0.5*9.8*0.1^2 = 0.049; dvz = 9.8*0.1 = 0.98
    delta << 0.01,0,0.049, 0,0,0,1, 0.2,0,0.98;
    ASSERT_MATRIX_APPROX(feat_imu->getMeasurement().transpose(), delta.transpose(), wolf::Constants::EPS);
}

//TEST_F(FeatureImu_test, addFactor)
//{
//    using namespace wolf;
//
//    FrameBasePtr frm_imu = std::static_pointer_cast<FrameBase>(last_frame);
//    auto cap_imu = last_frame->getCaptureList().back();
//    FactorImuPtr factor_imu = std::make_shared<FactorImu>(feat_imu, std::static_pointer_cast<CaptureImu>(cap_imu), processor_ptr_);
//    feat_imu->addFactor(factor_imu);
//    origin_frame->addConstrainedBy(factor_imu);
//}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
